@@ -6,17 +6,22 @@
  * exposes externally.
  *
  * To understand the structure of the RTS headers, see the wiki:
- *   http://ghc.haskell.org/trac/ghc/wiki/Commentary/SourceTree/Includes
+ *   https://gitlab.haskell.org/ghc/ghc/wikis/commentary/source-tree/includes
  *
  * ---------------------------------------------------------------------------*/
 
-#ifndef RTS_H
-#define RTS_H
+#pragma once
 
-#ifdef __cplusplus
+#if defined(__cplusplus)
 extern "C" {
 #endif
 
+/* get types from GHC's runtime system */
+#include "ghcconfig.h"
+/* We have to include Types.h before everything else as this defines some
+   macros that will change the behaviour of system headers.  */
+#include "stg/Types.h"
+
 /* We include windows.h very early, as on Win64 the CONTEXT type has
    fields "R8", "R9" and "R10", which goes bad if we've already
    #define'd those names for our own purposes (in stg/Regs.h) */
@@ -24,7 +29,7 @@ extern "C" {
 #include <windows.h>
 #endif
 
-#ifndef IN_STG_CODE
+#if !defined(IN_STG_CODE)
 #define IN_STG_CODE 0
 #endif
 #include "Stg.h"
@@ -32,12 +37,17 @@ extern "C" {
 #include "HsFFI.h"
 #include "RtsAPI.h"
 
-// Turn off inlining when debugging - it obfuscates things
-#ifdef DEBUG
+// Disencourage gcc from inlining when debugging - it obfuscates things
+#if defined(DEBUG)
 # undef  STATIC_INLINE
 # define STATIC_INLINE static
 #endif
 
+// Fine grained inlining control helpers.
+#define ATTR_ALWAYS_INLINE __attribute__((always_inline))
+#define ATTR_NOINLINE      __attribute__((noinline))
+
+
 #include "rts/Types.h"
 #include "rts/Time.h"
 
@@ -59,7 +69,13 @@ extern "C" {
 #if __GNUC__ >= 4
 #define RTS_UNLIKELY(p) __builtin_expect((p),0)
 #else
-#define RTS_UNLIKELY(p) p
+#define RTS_UNLIKELY(p) (p)
+#endif
+
+#if __GNUC__ >= 4
+#define RTS_LIKELY(p) __builtin_expect(!!(p), 1)
+#else
+#define RTS_LIKELY(p) (p)
 #endif
 
 /* __builtin_unreachable is supported since GNU C 4.5 */
@@ -69,14 +85,18 @@ extern "C" {
 #define RTS_UNREACHABLE abort()
 #endif
 
+/* Prefetch primitives */
+#define prefetchForRead(ptr) __builtin_prefetch(ptr, 0)
+#define prefetchForWrite(ptr) __builtin_prefetch(ptr, 1)
+
 /* Fix for mingw stat problem (done here so it's early enough) */
-#ifdef mingw32_HOST_OS
+#if defined(mingw32_HOST_OS)
 #define __MSVCRT__ 1
 #endif
 
 /* Needed to get the macro version of errno on some OSs, and also to
    get prototypes for the _r versions of C library functions. */
-#ifndef _REENTRANT
+#if !defined(_REENTRANT)
 #define _REENTRANT 1
 #endif
 
@@ -98,24 +118,36 @@ extern "C" {
 void _assertFail(const char *filename, unsigned int linenum)
    GNUC3_ATTRIBUTE(__noreturn__);
 
+void _warnFail(const char *filename, unsigned int linenum);
+
 #define CHECK(predicate)                        \
-        if (predicate)                          \
+        if (RTS_LIKELY(predicate))              \
             /*null*/;                           \
         else                                    \
             _assertFail(__FILE__, __LINE__)
 
+#define CHECKWARN(predicate)                        \
+        if (RTS_LIKELY(predicate))              \
+            /*null*/;                           \
+        else                                    \
+            _warnFail(__FILE__, __LINE__)
+
 #define CHECKM(predicate, msg, ...)             \
-        if (predicate)                          \
+        if (RTS_LIKELY(predicate))              \
             /*null*/;                           \
         else                                    \
             barf(msg, ##__VA_ARGS__)
 
-#ifndef DEBUG
+#if !defined(DEBUG)
 #define ASSERT(predicate) /* nothing */
 #define ASSERTM(predicate,msg,...) /* nothing */
+#define WARN(predicate) /* nothing */
 #else
 #define ASSERT(predicate) CHECK(predicate)
 #define ASSERTM(predicate,msg,...) CHECKM(predicate,msg,##__VA_ARGS__)
+#define WARN(predicate) CHECKWARN(predicate)
+#undef ASSERTS_ENABLED
+
 #endif /* DEBUG */
 
 /*
@@ -126,7 +158,7 @@ void _assertFail(const char *filename, unsigned int linenum)
 
 #define doNothing() do { } while (0)
 
-#ifdef DEBUG
+#if defined(DEBUG)
 #define USED_IF_DEBUG
 #define USED_IF_NOT_DEBUG STG_UNUSED
 #else
@@ -134,7 +166,7 @@ void _assertFail(const char *filename, unsigned int linenum)
 #define USED_IF_NOT_DEBUG
 #endif
 
-#ifdef THREADED_RTS
+#if defined(THREADED_RTS)
 #define USED_IF_THREADS
 #define USED_IF_NOT_THREADS STG_UNUSED
 #else
@@ -142,6 +174,14 @@ void _assertFail(const char *filename, unsigned int linenum)
 #define USED_IF_NOT_THREADS
 #endif
 
+#if defined(PROFILING)
+#define USED_IF_PROFILING
+#define USED_IF_NOT_PROFILING STG_UNUSED
+#else
+#define USED_IF_PROFILING STG_UNUSED
+#define USED_IF_NOT_PROFILING
+#endif
+
 #define FMT_SizeT    "zu"
 #define FMT_HexSizeT "zx"
 
@@ -161,12 +201,17 @@ void _assertFail(const char *filename, unsigned int linenum)
 /* Global constraints */
 #include "rts/Constants.h"
 
+/* Runtime flags */
+#include "rts/Flags.h"
+
 /* Profiling information */
 #include "rts/prof/CCS.h"
+#include "rts/prof/Heap.h"
 #include "rts/prof/LDV.h"
 
 /* Parallel information */
 #include "rts/OSThreads.h"
+#include "rts/TSANUtils.h"
 #include "rts/SpinLock.h"
 
 #include "rts/Messages.h"
@@ -176,6 +221,7 @@ void _assertFail(const char *filename, unsigned int linenum)
 #include "rts/storage/FunTypes.h"
 #include "rts/storage/InfoTables.h"
 #include "rts/storage/Closures.h"
+#include "rts/storage/Heap.h"
 #include "rts/storage/ClosureTypes.h"
 #include "rts/storage/TSO.h"
 #include "stg/MiscClosures.h" /* InfoTables, closures etc. defined in the RTS */
@@ -183,26 +229,33 @@ void _assertFail(const char *filename, unsigned int linenum)
 #include "rts/storage/ClosureMacros.h"
 #include "rts/storage/MBlock.h"
 #include "rts/storage/GC.h"
+#include "rts/NonMoving.h"
+
+/* Foreign exports */
+#include "rts/ForeignExports.h"
 
 /* Other RTS external APIs */
+#include "rts/ExecPage.h"
 #include "rts/Parallel.h"
 #include "rts/Signals.h"
 #include "rts/BlockSignals.h"
 #include "rts/Hpc.h"
-#include "rts/Flags.h"
 #include "rts/Adjustor.h"
 #include "rts/FileLock.h"
 #include "rts/GetTime.h"
 #include "rts/Globals.h"
-#include "rts/IOManager.h"
+#include "rts/IOInterface.h"
 #include "rts/Linker.h"
 #include "rts/Ticky.h"
 #include "rts/Timer.h"
-#include "rts/Stable.h"
+#include "rts/StablePtr.h"
+#include "rts/StableName.h"
 #include "rts/TTY.h"
 #include "rts/Utils.h"
 #include "rts/PrimFloat.h"
 #include "rts/Main.h"
+#include "rts/Profiling.h"
+#include "rts/IPE.h"
 #include "rts/StaticPtrTable.h"
 #include "rts/Libdw.h"
 #include "rts/LibdwPool.h"
@@ -212,18 +265,12 @@ DLL_IMPORT_RTS extern char **prog_argv; /* so we can get at these from Haskell *
 DLL_IMPORT_RTS extern int    prog_argc;
 DLL_IMPORT_RTS extern char  *prog_name;
 
-#ifdef mingw32_HOST_OS
-// We need these two from Haskell too
-void getWin32ProgArgv(int *argc, wchar_t **argv[]);
-void setWin32ProgArgv(int argc, wchar_t *argv[]);
-#endif
-
 void reportStackOverflow(StgTSO* tso);
 void reportHeapOverflow(void);
 
 void stg_exit(int n) GNU_ATTRIBUTE(__noreturn__);
 
-#ifndef mingw32_HOST_OS
+#if !defined(mingw32_HOST_OS)
 int stg_sig_install (int, int, void *);
 #endif
 
@@ -237,6 +284,15 @@ int rts_isProfiled(void);
 // Returns non-zero if the RTS is a dynamically-linked version
 int rts_isDynamic(void);
 
+// Returns non-zero if the RTS is a threaded version
+int rts_isThreaded(void);
+
+// Returns non-zero if the RTS is a debugged version
+int rts_isDebugged(void);
+
+// Returns non-zero if the RTS is a tracing version (event log)
+int rts_isTracing(void);
+
 /* -----------------------------------------------------------------------------
    RTS Exit codes
    -------------------------------------------------------------------------- */
@@ -252,7 +308,7 @@ int rts_isDynamic(void);
    Miscellaneous garbage
    -------------------------------------------------------------------------- */
 
-#ifdef DEBUG
+#if defined(DEBUG)
 #define TICK_VAR(arity) \
   extern StgInt SLOW_CALLS_##arity; \
   extern StgInt RIGHT_ARITY_##arity; \
@@ -270,27 +326,28 @@ TICK_VAR(2)
 
 #define IF_RTSFLAGS(c,s)  if (RtsFlags.c) { s; } doNothing()
 
-#ifdef DEBUG
+#if defined(DEBUG)
+/* See Note [RtsFlags is a pointer in STG code] */
 #if IN_STG_CODE
 #define IF_DEBUG(c,s)  if (RtsFlags[0].DebugFlags.c) { s; } doNothing()
 #else
 #define IF_DEBUG(c,s)  if (RtsFlags.DebugFlags.c) { s; } doNothing()
-#endif
+#endif /* IN_STG_CODE */
 #else
 #define IF_DEBUG(c,s)  doNothing()
-#endif
+#endif /* DEBUG */
 
-#ifdef DEBUG
+#if defined(DEBUG)
 #define DEBUG_ONLY(s) s
 #else
 #define DEBUG_ONLY(s) doNothing()
-#endif
+#endif /* DEBUG */
 
-#ifdef DEBUG
+#if defined(DEBUG)
 #define DEBUG_IS_ON   1
 #else
 #define DEBUG_IS_ON   0
-#endif
+#endif /* DEBUG */
 
 /* -----------------------------------------------------------------------------
    Useful macros and inline functions
@@ -310,8 +367,6 @@ TICK_VAR(2)
 
 /* -------------------------------------------------------------------------- */
 
-#ifdef __cplusplus
+#if defined(__cplusplus)
 }
 #endif
-
-#endif /* RTS_H */
